# types.py
# Copyright (C) 2012 the ColanderAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of ColanderAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php
try:
    from collections import OrderedDict

except ImportError:
    from backports import OrderedDict

from sqlalchemy.orm.properties import ColumnProperty
from sqlalchemy.orm.properties import RelationshipProperty
from sqlalchemy.orm import class_mapper

__all__ = ['MappingRegistry']


class MappingRegistry(object):

    def __init__(self, cls, excludes=None, includes=None, nullables=None):
        """ Keep information about the SQLAlchemy mapped class `cls`.
        """
        self.cls = cls
        self._mapper = class_mapper(cls)
        self.excludes = excludes or {}
        self.includes = includes or {}
        self.nullables = nullables or {}
        self.pkeys = [col.name for col in self._mapper.primary_key]
        self.fkeys = {}
        self.rkeys = {}
        self.attrs = {}
        self.properties = {}
        self.fields = set()
        self.relationships = set()
        self.references = set()
        self.collections = set()

        for p in self._mapper.iterate_properties:

            self.properties[p.key] = p

            if isinstance(p, ColumnProperty):

                col = p.columns[0]
                self.attrs[p.key] = col
                self.fields.add(p.key)

                reg = col.ca_registry.copy() if hasattr(col, 'ca_registry') else {}
                if p.key not in self.includes and\
                   'include' in reg and reg['include']:
                    self.includes[p.key] = reg['include']

                if p.key not in self.excludes and\
                   'exclude' in reg and reg['exclude']:
                    self.excludes[p.key] = reg['exclude']

                if p.key not in self.nullables and 'nullable' in reg:
                    self.nullables[p.key] = reg['nullable']

            elif isinstance(p, RelationshipProperty):

                if callable(p.argument):
                    cls = p.argument()
                else:
                    cls = p.argument.class_

                reg = p.ca_registry if hasattr(p, 'ca_registry') else {}
                if p.key not in self.includes and\
                   'include' in reg and reg['include']:
                    self.includes[p.key] = reg['include']

                if p.key not in self.excludes and\
                   'exclude' in reg and reg['exclude']:
                    self.excludes[p.key] = reg['exclude']

                if p.key not in self.nullables and 'nullable' in reg:
                    self.nullables[p.key] = reg['nullable']

                self.attrs[p.key] = cls
                self.relationships.add(p.key)
                self.rkeys[p.key] = [col.name
                                     for col in class_mapper(cls).primary_key]
                if p.uselist:
                    self.collections.add(p.key)
                else:
                    self.references.add(p.key)
                    self.fkeys[p.key] = OrderedDict()
                    for col in p._calculated_foreign_keys:
                        if col.table in self._mapper.tables:
                            for f in col.foreign_keys:
                                self.fkeys[p.key][col.name] = f.column.name

                    if not self.fkeys[p.key]:
                        self.fkeys.pop(p.key)
            else:
                msg = 'Unsupported property type: {}'.format(type(p))
                raise NotImplementedError(msg)

            if p.key in self.includes and \
               p.key in self.excludes and \
               not self.includes[p.key] is None and \
               not self.excludes[p.key] is None and \
               self.includes[p.key] == self.excludes[p.key]:
                msg = "%s cannot be included and excluded at the same time."
                raise ValueError(msg % p.key)



# __init__.py
# Copyright (C) 2012 the ColanderAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of ColanderAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

import test_declarative
import test_types

__all__ = ['test_declarative', 'test_types']



# types.py
# Copyright (C) 2012 the ColanderAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of ColanderAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from sqlalchemy.orm import class_mapper
from sqlalchemy.orm import object_mapper
from .utils import MappingRegistry
import colander
import sqlalchemy.types

__all__ = ['SQLAlchemyMapping']


class SQLAlchemyMapping(colander.SchemaNode):

    def __init__(self, cls, excludes=None,
                 includes=None, nullables=None, unknown='raise'):
        """ Build a Colander Schema based on the SQLAlchemy mapped class.
        """
        super(SQLAlchemyMapping, self).__init__(colander.Mapping(unknown))

        self._reg = MappingRegistry(cls, excludes, includes, nullables)
        for name, obj in self._reg.attrs.items():
            if name in self._reg.excludes:
                continue

            if self._reg.includes and name not in self._reg.includes:
                continue

            if name in self._reg.fields:
                node = self.get_schema_from_col(obj,
                                                self._reg.nullables.get(name))

            else:
                node = self.get_schema_from_rel(obj,
                                                name,
                                                name in self._reg.collections)
            self.add(node)

    @property
    def registry(self):
        return self._reg

    def get_schema_from_col(self, column, nullable=None):
        """ Build and return a Colander SchemaNode
            using information stored in the column.
        """

        if hasattr(column, 'ca_registry'):
            params = column.ca_registry.copy()

        else:
            params = {}

        # support sqlalchemy.types.TypeDecorator
        column_type = getattr(column.type, 'impl', column.type)

        if 'type' in params:
            type_ = params.pop('type')

        elif isinstance(column_type, sqlalchemy.types.Boolean):
            type_ = colander.Boolean()

        elif isinstance(column_type, sqlalchemy.types.Date):
            type_ = colander.Date()

        elif isinstance(column_type, sqlalchemy.types.DateTime):
            type_ = colander.DateTime()

        elif isinstance(column_type, sqlalchemy.types.Enum):
            type_ = colander.String()

        elif isinstance(column_type, sqlalchemy.types.Float):
            type_ = colander.Float()

        elif isinstance(column_type, sqlalchemy.types.Integer):
            type_ = colander.Integer()

        elif isinstance(column_type, sqlalchemy.types.String):
            type_ = colander.String()

        elif isinstance(column_type, sqlalchemy.types.Numeric):
            type_ = colander.Decimal()

        elif isinstance(column_type, sqlalchemy.types.Time):
            type_ = colander.Time()

        else:
            raise NotImplementedError('Unknown type: %s' % column.type)

        if 'children' in params:
            children = params.pop('children')
        else:
            children = []

        # Add a default value for missing parameters during serialization.
        if 'default' not in params and column.default is None:
            params['default'] = colander.null

        elif 'default' not in params and not column.default is None:
            if column.default.is_callable:
                # Fix: SQLA wraps callables in lambda ctx: fn().
                default = column.default.arg(None)
            else:
                default = column.default.arg
            params['default'] = default

        # Add a default value for  missing parameters during deserialization.
        if 'missing' not in params and not column.nullable:
            params['missing'] = colander.required

        elif 'missing' not in params and column.default is None:
            params['missing'] = None

        elif 'missing' not in params and not column.default is None:
            if column.default.is_callable:
                # Fix: SQLA wraps default callables in lambda ctx: fn().
                default = column.default.arg(None)
            else:
                default = column.default.arg
            params['missing'] = default

        # Overwrite default missing value when nullable is specified.
        if nullable is False:
            params['missing'] = colander.required

        elif nullable is True:
            params['missing'] = None

        if 'validator' not in params and \
           isinstance(column_type, sqlalchemy.types.Enum):

            params['validator'] = colander.OneOf(column.type.enums)

        elif 'validator' not in params and \
           isinstance(column_type, sqlalchemy.types.Enum):

            params['validator'] = colander.Length(0, column.type.length)

        # The name of SchemaNode must be the same of SQLA class attribute.
        params['name'] = column.name

        return colander.SchemaNode(type_, *children, **params)

    def get_schema_from_rel(self, cls, name, uselist=False, nullable=None):
        """ Build and return a Colander SchemaNode
            using information stored in the relationship property.
        """

        if hasattr(self._reg.properties[name], 'ca_registry'):
            params = self._reg.properties[name].ca_registry.copy()

        else:
            params = {}

        # The name of SchemaNode must be the same of SQLA class attribute.
        params['name'] = name

        if 'type' in params:
            type_ = params.pop('type')

        elif uselist:
            # xToMany relationships.
            type_ = colander.Sequence()

        else:
            # xToOne relationships.
            type_ = colander.Mapping()

        if 'children' in params:
            children = params.pop('children')

        else:
            mapper = class_mapper(cls)
            children = [self.get_schema_from_col(col)
                        for col in mapper.primary_key]

        if nullable == False:
            params['missing'] = colander.required

        elif 'type' not in params and uselist:
            params['missing'] = []

        elif 'type' not in params:
            params['missing'] = None

        if 'default' not in params:
            params['default'] = colander.null

        return colander.SchemaNode(type_, *children, **params)

    def dictify(self, obj):
        """ Build and return a dictified version of `obj`
            using schema information to choose what attributes
            will be included in the returned dict.
        """
        dict_ = {}
        for name in self._reg.attrs:

            if name in self._reg.excludes or\
              (self._reg.includes and name not in self._reg.includes):
                continue

            if name in self._reg.fields:
                dict_[name] = getattr(obj, name)

            elif name in self._reg.references:
                value = getattr(obj, name)
                if not value is None:
                    value = self.dictify_relationship(value)
                dict_[name] = value

            elif name in self._reg.collections:
                dict_[name] = [self.dictify_relationship(value)
                               for value in getattr(obj, name)]

        return dict_

    def dictify_relationship(self, obj):
        dict_ = {}
        for col in object_mapper(obj).primary_key:
            dict_[col.name] = getattr(obj, col.name)
        return dict_

    def clone(self):
        cloned = self.__class__(self._reg.cls,
                                self._reg.excludes,
                                self._reg.includes,
                                self._reg.nullables)
        cloned.__dict__.update(self.__dict__)
        cloned.children = [node.clone() for node in self.children]
        return cloned



# test_declarative.py
# Copyright (C) 2012 the ColanderAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of ColanderAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from models import Account
from models import Address
from models import Contact
from models import LogEntry
from models import Person
from models import Template
from models import Theme
import colander
import colanderalchemy
import datetime
import logging
import sys

if sys.version_info[0] == 2 and sys.version_info[1] < 7:
    # In Python < 2.7 use unittest2.
    import unittest2 as unittest

else:
    import unittest


log = logging.getLogger(__name__)


class TestsSQLAlchemyMapping(unittest.TestCase):

    def setUp(self):
        self.account = colanderalchemy.SQLAlchemyMapping(Account)
        self.address = colanderalchemy.SQLAlchemyMapping(Address)
        self.contact = colanderalchemy.SQLAlchemyMapping(Contact)
        self.logentry = colanderalchemy.SQLAlchemyMapping(LogEntry)
        self.person = colanderalchemy.SQLAlchemyMapping(Person)
        self.template = colanderalchemy.SQLAlchemyMapping(Template)
        self.theme = colanderalchemy.SQLAlchemyMapping(Theme)

    def tearDown(self):
        pass

    def test_account(self):
        # Test for required fields (ex. primary keys).
        self.assertRaises(colander.Invalid, self.account.deserialize, {})
        # Test below fails due to nullable fields.
        self.assertRaises(colander.Invalid, self.account.deserialize,
                          {'email': 'mailbox@domain.tld'})
        self.assertRaises(colander.Invalid, self.account.deserialize,
                          {'email': 'mailbox@domain.tld',
                           'name': 'My Name.'})
        self.assertRaises(colander.Invalid, self.account.deserialize,
                          {'email': 'mailbox@domain.tld',
                           'name': 'My Name.',
                           'surname': 'My Surname.'})
        self.assertRaises(colander.Invalid, self.account.deserialize,
                          {'email': 'mailbox@domain.tld',
                           'name': 'My Name.',
                           'surname': 'My Surname.',
                           'gender': 'A'})

    def test_relationship(self):
        self.assertEqual(self.account['contact'].title, "Contact")
        self.assertEqual(self.account['contact'].name, "contact")

    def test_account_registry(self):
        self.assertIn('email', self.account._reg.pkeys)
        self.assertIn('email', self.account._reg.fields)
        self.assertIn('name', self.account._reg.fields)
        self.assertIn('surname', self.account._reg.fields)
        self.assertIn('gender', self.account._reg.fields)
        self.assertIn('themes', self.account._reg.relationships)
        self.assertIn('themes', self.account._reg.rkeys)
        self.assertIn('contact', self.account._reg.references)
        self.assertEqual(self.account._reg.fkeys, {})
        self.assertIn('themes', self.account._reg.collections)

    def test_contact_registry(self):
        self.assertIn('type_', self.contact._reg.pkeys)
        self.assertIn('account_id', self.contact._reg.pkeys)
        self.assertIn('type_', self.contact._reg.fields)
        self.assertIn('value', self.contact._reg.fields)
        self.assertIn('account_id', self.contact._reg.fields)
        self.assertIn('account', self.contact._reg.relationships)
        self.assertIn('account', self.contact._reg.rkeys)
        self.assertIn('account', self.contact._reg.references)
        self.assertIn('account', self.contact._reg.fkeys)
        self.assertEqual(self.contact._reg.collections, set())

    def test_theme(self):
        self.assertRaises(colander.Invalid, self.theme.deserialize, {})
        self.assertRaises(colander.Invalid, self.theme.deserialize,
                          {'name': 'My Name.'})
        self.assertRaises(colander.Invalid, self.theme.deserialize,
                          {'author_id': 'mailbox@domain.tld'})
        self.assertRaises(colander.Invalid, self.theme.deserialize,
                          {'name': 'My Name.',
                           'author_id': 'mailbox@domain.tld',
                           'author': None})
        self.assertRaises(colander.Invalid, self.theme.deserialize,
                          {'name': 'My Name.',
                           'author_id': 'mailbox@domain.tld',
                           'author': {}})

    def test_theme_registry(self):
        self.assertIn('name', self.theme._reg.pkeys)
        self.assertIn('author_id', self.theme._reg.pkeys)
        self.assertIn('name', self.theme._reg.fields)
        self.assertIn('description', self.theme._reg.fields)
        self.assertIn('author_id', self.theme._reg.fields)
        self.assertIn('author', self.theme._reg.relationships)
        self.assertIn('author', self.theme._reg.rkeys)
        self.assertIn('author', self.theme._reg.references)
        self.assertIn('author', self.theme._reg.fkeys)
        self.assertIn('templates', self.theme._reg.collections)

    def test_template(self):
        self.assertRaises(colander.Invalid, self.template.deserialize, {})
        self.assertRaises(colander.Invalid, self.template.deserialize,
                          {'name': 'My Name.'})
        self.assertRaises(colander.Invalid, self.template.deserialize,
                          {'theme_name': 'Its Name.'})
        self.assertRaises(colander.Invalid, self.template.deserialize,
                          {'theme_author_id': 'mailbox@domain.tld'})
        self.assertRaises(colander.Invalid, self.template.deserialize,
                          {'name': 'My Name.',
                           'theme_name': 'Its Name.',
                           'theme_author_id': 'mailbox@domain.tld',
                           'theme': {}})
        self.assertRaises(colander.Invalid, self.template.deserialize,
                          {'name': 'My Name.',
                           'theme_name': 'Its Name.',
                           'theme_author_id': 'mailbox@domain.tld',
                           'theme': None})

    def test_template_registry(self):
        self.assertIn('name', self.template._reg.pkeys)
        self.assertIn('theme_name', self.template._reg.pkeys)
        self.assertIn('theme_author_id', self.template._reg.pkeys)
        self.assertIn('name', self.template._reg.fields)
        self.assertIn('theme_name', self.template._reg.fields)
        self.assertIn('theme_author_id', self.template._reg.fields)
        self.assertIn('theme', self.template._reg.relationships)
        self.assertIn('theme', self.template._reg.rkeys)
        self.assertIn('theme', self.template._reg.references)
        self.assertIn('theme', self.template._reg.fkeys)
        self.assertEqual(self.template._reg.collections, set())

    def test_excludes(self):
        excludes = {
            'email': True,
            'name': True,
            'surname': True,
            'gender': True,
            'contact': True,
            'themes': True
        }
        account = colanderalchemy.SQLAlchemyMapping(Account, excludes=excludes)
        data = account.deserialize({})
        self.assertEqual(data, {})
        self.assertEqual(account.serialize(data), {})

    def test_includes(self):
        includes = {'email': True}
        account = colanderalchemy.SQLAlchemyMapping(Account, includes=includes)
        self.assertEqual(list(account.serialize({}).keys()), ['email'])

        self.assertRaises(ValueError, colanderalchemy.SQLAlchemyMapping,
                          Account, {'email': True}, {'email': True})
        self.assertRaises(ValueError, colanderalchemy.SQLAlchemyMapping,
                          Account, {'email': False}, {'email': False})

    def test_nullables(self):
        nullables = {
            'email': True,
            'name': True,
            'surname': True,
            'gender': True,
            'contact': True,
            'themes': True
        }
        account = colanderalchemy.SQLAlchemyMapping(Account,
                                                    nullables=nullables)
        data = {
            'contact': None,
            'email': None,
            'gender': None,
            'name': None,
            'surname': None,
            'themes': []
        }
        self.assertEqual(data, account.deserialize({}))
        self.assertEqual(account.serialize({}), {
            'contact': {'account_id': colander.null, 'type_': colander.null},
            'email': colander.null,
            'gender': colander.null,
            'name': colander.null,
            'surname': colander.null,
            'themes': colander.null
        })

    def test_clone(self):
        account = colanderalchemy.SQLAlchemyMapping(Account)
        account2 = account.clone()
        self.assertEqual(account._reg, account2._reg)
        self.assertEqual(len(account.children), len(account2.children))

    def test_dictify(self):
        params = {'email': 'mailbox@domain.tld',
                  'name': 'My Name.',
                  'surname': 'My Surname.',
                  'gender': 'M'}
        account = Account(**self.account.deserialize(params))
        dictified = self.account.dictify(account)
        params = {'email': 'mailbox@domain.tld',
                  'name': 'My Name.',
                  'surname': 'My Surname.',
                  'gender': 'M',
                  'contact': None,
                  'themes': []}
        self.assertEqual(dictified, params)

    def test_datetime(self):
        params = self.logentry.deserialize({'id': 1})
        self.assertIn('created', params)
        self.assertIn('modified', params)
        self.assertEqual(isinstance(params['created'], datetime.datetime),
                         True)
        self.assertEqual(isinstance(params['modified'], datetime.datetime),
                         True)



# __init__.py
# Copyright (C) 2012 the ColanderAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of ColanderAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from .types import SQLAlchemyMapping
from .utils import MappingRegistry
from .declarative import Column
from .declarative import relationship

__all__ = ['SQLAlchemyMapping', 'MappingRegistry', 'Column', 'relationship']



# test_declarative.py
# Copyright (C) 2012 the ColanderAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of ColanderAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from models import Address
from models import Person
from models import Versioned
import colander
import colanderalchemy
import logging
import sys

if sys.version_info[0] == 2 and sys.version_info[1] < 7:
    # In Python < 2.7 use unittest2.
    import unittest2 as unittest

else:
    import unittest


log = logging.getLogger(__name__)


class TestsBase(unittest.TestCase):

    def setUp(self):
        self.address = colanderalchemy.SQLAlchemyMapping(Address)
        self.person = colanderalchemy.SQLAlchemyMapping(Person)
        self.versioned = colanderalchemy.SQLAlchemyMapping(Versioned)

    def tearDown(self):
        pass

    def test_column(self):
        self.assertEqual(type(self.person['id'].typ), colander.Float)
        self.assertEqual(self.person['id'].title, 'Person ID')
        self.assertEqual(self.person['id'].description, 'The Person id.')
        self.assertEqual(self.person['id'].widget, 'Empty Widget')
        self.assertRaises(KeyError, self.person.__getitem__, 'surname')
        self.assertEqual(self.person['name'].default, colander.required)
        self.assertEqual(self.address['id'].missing, colander.null)

    def test_relationship(self):
        self.assertRaises(KeyError, self.person.__getitem__, 'addresses')
        self.assertEqual(self.address['person'].title, 'Person relationship.')
        self.assertEqual(self.address['person'].description, 'The Person desc.')
        self.assertEqual(self.address['person'].widget, 'Rel. Empty Widget')

    def test_column_inheritance(self):
        self.assertRaises(colander.Invalid,
                          self.versioned.deserialize,
                          {})
        self.assertRaises(colander.Invalid,
                          self.versioned.deserialize,
                          {'id': '1'})



from setuptools import setup, find_packages

version = '0.1'

setup(name='ColanderAlchemy',
      version=version,
      description="Autogenerate Colander schemas based on SQLAlchemy models.",
      long_description="""\
An helper tool for autogenerating Colander schemas based on SQLAlchemy mapped classes.""",
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 2.6',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.2',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Topic :: Database'],
      keywords='serialize deserialize validate schema validation colander sqlalchemy',
      author='Stefano Fontanelli',
      author_email='s.fontanelli@asidev.com',
      url='https://github.com/stefanofontanelli/ColanderAlchemy',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
            'colander',
            'SQLAlchemy'],
      tests_require=[],
      test_suite='tests',
      )



# declarative.py
# Copyright (C) 2012 the ColanderAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of ColanderAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

import sqlalchemy.schema
import sqlalchemy.ext.declarative

__all__ = ['Column', 'relationship']


class Column(sqlalchemy.schema.Column):

    def __init__(self, *args, **kwargs):

        self._ca_registry = {}
        for key in ['ca_type', 'ca_children', 'ca_default',
                    'ca_missing', 'ca_preparer', 'ca_validator',
                    'ca_after_bind', 'ca_title', 'ca_description',
                    'ca_widget', 'ca_include', 'ca_exclude', 'ca_nullable']:
            try:
                value = kwargs.pop(key)

            except KeyError:
                continue

            else:
                self._ca_registry[key[3:]] = value

        super(Column, self).__init__(*args, **kwargs)

    @property
    def ca_registry(self):
        return self._ca_registry

    def copy(self, **kwargs):
        col = super(Column, self).copy(**kwargs)
        col._ca_registry = self._ca_registry.copy()
        return col


def relationship(argument, secondary=None, **kwargs):

    registry = {}
    for key in ['ca_type', 'ca_children', 'ca_default',
                'ca_missing', 'ca_preparer', 'ca_validator', 'ca_after_bind',
                'ca_title', 'ca_description', 'ca_widget',
                'ca_include', 'ca_exclude', 'ca_nullable']:
        try:
            value = kwargs.pop(key)

        except KeyError:
            continue

        else:
            registry[key[3:]] = value

    relationship = sqlalchemy.orm.relationship(argument, secondary, **kwargs)
    relationship.ca_registry = registry
    return relationship



# models.py
# Copyright (C) 2012 the ColanderAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of ColanderAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

import colander
import colanderalchemy
import datetime
import sqlalchemy
import sqlalchemy.ext.declarative
import sqlalchemy.orm
import sqlalchemy.schema
import sys


Base = sqlalchemy.ext.declarative.declarative_base()


class Account(Base):

    __tablename__ = 'accounts'

    email = sqlalchemy.Column(sqlalchemy.Unicode(256),
                              primary_key=True)
    name = sqlalchemy.Column(sqlalchemy.Unicode(128),
                             nullable=False)
    surname = sqlalchemy.Column(sqlalchemy.Unicode(128),
                                nullable=False)
    gender = sqlalchemy.Column(sqlalchemy.Enum('M', 'F'),
                               nullable=False)
    contact = sqlalchemy.orm.relationship('Contact',
                                          uselist=False,
                                          back_populates='account')
    themes = sqlalchemy.orm.relationship('Theme',
                                         back_populates='author')


class Contact(Base):

    __tablename__ = 'contacts'

    type_ = sqlalchemy.Column(sqlalchemy.Unicode(256),
                              primary_key=True)
    value = sqlalchemy.Column(sqlalchemy.Unicode(256),
                              nullable=False)
    account_id = sqlalchemy.Column(sqlalchemy.Unicode(256),
                                   sqlalchemy.ForeignKey('accounts.email'),
                                   primary_key=True)
    account = sqlalchemy.orm.relationship('Account',
                                          back_populates='contact')


class Theme(Base):

    __tablename__ = 'themes'

    name = sqlalchemy.Column(sqlalchemy.Unicode(256),
                             primary_key=True)
    description = sqlalchemy.Column(sqlalchemy.UnicodeText,
                                    default='')
    author_id = sqlalchemy.Column(sqlalchemy.Unicode(256),
                                  sqlalchemy.ForeignKey('accounts.email'),
                                  primary_key=True)
    author = sqlalchemy.orm.relationship('Account',
                                         back_populates='themes')
    templates = sqlalchemy.orm.relationship('Template',
                                            back_populates='theme')


class Template(Base):

    __tablename__ = 'templates'

    __table_args__ = (
        sqlalchemy.schema.ForeignKeyConstraint(
            ['theme_name', 'theme_author_id'],
            ['themes.name', 'themes.author_id'],
            onupdate='CASCADE',
            ondelete='CASCADE'
        ),
    )

    name = sqlalchemy.Column(sqlalchemy.Unicode(128),
                             primary_key=True)
    theme_name = sqlalchemy.Column(sqlalchemy.Unicode(256),
                                   primary_key=True)
    theme_author_id = sqlalchemy.Column(sqlalchemy.Unicode(256),
                                        primary_key=True)
    theme = sqlalchemy.orm.relationship('Theme',
                                        back_populates='templates')


class Person(Base):

    __tablename__ = 'person'
    id = colanderalchemy.Column(sqlalchemy.Integer,
                                primary_key=True,
                                ca_type=colander.Float(),
                                ca_title='Person ID',
                                ca_description='The Person id.',
                                ca_widget='Empty Widget',
                                ca_include=True)
    name = colanderalchemy.Column(sqlalchemy.Unicode(128),
                                  nullable=False,
                                  ca_nullable=True,
                                  ca_include=True,
                                  ca_default=colander.required)
    surname = colanderalchemy.Column(sqlalchemy.Unicode(128),
                                     nullable=False,
                                     ca_exclude=True)
    addresses = colanderalchemy.relationship('Address',
                                             ca_exclude=True)


class Address(Base):

    __tablename__ = 'addresses'

    id = colanderalchemy.Column(sqlalchemy.Integer,
                                primary_key=True,
                                ca_missing=colander.null)
    street = colanderalchemy.Column(sqlalchemy.Unicode(128),
                                    nullable=False)
    city = colanderalchemy.Column(sqlalchemy.Unicode(128),
                                  nullable=False)
    person_id = colanderalchemy.Column(sqlalchemy.Unicode(128),
                                       sqlalchemy.ForeignKey('person.id'),
                                       primary_key=True)
    person = colanderalchemy.relationship('Person',
                                          ca_title='Person relationship.',
                                          ca_description='The Person desc.',
                                          ca_widget='Rel. Empty Widget')


class LogEntry(Base):

    __tablename__ = 'logentries'

    id = sqlalchemy.Column(sqlalchemy.Integer,
                           primary_key=True)
    created = sqlalchemy.Column(sqlalchemy.DateTime,
                                default=datetime.datetime.utcnow,
                                nullable=True)
    modified = sqlalchemy.Column(sqlalchemy.DateTime,
                                 default=lambda: datetime.datetime.utcnow() + \
                                                 datetime.timedelta(1),
                                 nullable=True)


class Timestamped(object):
    """ An automatically timestamped mixin. """
    ins_date = colanderalchemy.Column(sqlalchemy.DateTime,
                                      nullable=False,
                                      default=datetime.datetime.utcnow,
                                      ca_exclude=True)
    mod_date = colanderalchemy.Column(sqlalchemy.DateTime,
                                      ca_include=False,
                                      nullable=False,
                                      default=datetime.datetime.utcnow)


TimeBase = sqlalchemy.ext.declarative.declarative_base(cls=Timestamped)


class Versioned(TimeBase):

    __tablename__ = 'versions'
    id = colanderalchemy.Column(sqlalchemy.Integer, primary_key=True)
